import pyro
import pyro.distributions as dist
from pyro.optim import ExponentialLR
from pyro.infer import SVI, JitTraceEnum_ELBO, TraceEnum_ELBO, config_enumerate

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.distributions.utils import logits_to_probs, probs_to_logits, clamp_probs
from torch.distributions import constraints
from torch.distributions.transforms import SoftmaxTransform

from .utils.custom_mlp import MLP, Exp, ZeroBiasMLP2
from .utils.utils import CustomDataset, CustomDataset2, CustomDataset4, tensor_to_numpy, convert_to_tensor

from .dist.negbinomial import NegativeBinomial as MyNB
from .dist.negbinomial import ZeroInflatedNegativeBinomial as MyZINB
from .graph.graph_utils import ClusterNetworkBuilder, KnnClusterNetworkBuilder,visualize_network,keep_top_k_edges_per_node,visualize_network_interactive,EdgeAllocation

from .SURE_nsf import SURENF
from .SURE_vae import SUREVAE
from .SURE_vanilla import SUREVanilla

import zuko 
from pyro.contrib.zuko import ZukoToPyro

import os
import argparse
import random
import numpy as np
import datatable as dt
from tqdm import tqdm
from scipy import sparse
from scipy.stats import epps_singleton_2samp, ks_2samp
import networkx as nx
import pandas as pd 
import plotly.graph_objects as go
import plotly.express as px
from sklearn.decomposition import PCA

import scanpy as sc
from .atac import binarize

from typing import Literal

import warnings
warnings.filterwarnings("ignore")

import dill as pickle
import gzip
from packaging.version import Version
torch_version = torch.__version__


def set_random_seed(seed):
    # Set seed for PyTorch
    torch.manual_seed(seed)
    
    # If using CUDA, set the seed for CUDA
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)  # For multi-GPU setups.
    
    # Set seed for NumPy
    np.random.seed(seed)
    
    # Set seed for Python's random module
    random.seed(seed)

    # Set seed for Pyro
    pyro.set_rng_seed(seed)

class SURE(nn.Module):
    """SUccinct REpresentation of single-omics cells

    Parameters
    ----------
    inpute_size
        Number of features (e.g., genes, peaks, proteins, etc.) per cell.
    codebook_size
        Number of metacells.
    covariate_size
        Number of cell-level factors. 
    transforms
        Number of neural spline flows
    z_dim
        Dimensionality of latent states and metacells. 
    hidden_layers
        A list gives the numbers of neurons for each hidden layer.
    flow_hidden_layers
        A list gives the numbers of neurons for each hidden layer in neural spline flows.
    loss_func
        The likelihood model for single-cell data generation. 
        
        One of the following: 
        * ``'negbinomial'`` -  negative binomial distribution (default)
        * ``'poisson'`` - poisson distribution
        * ``'multinomial'`` - multinomial distribution
    use_cuda
        A boolean option for switching on cuda device. 

    Examples
    --------
    >>>
    >>>
    >>>

    """
    def __init__(self,
                 input_dim: int,
                 codebook_size: int,
                 condition_size: int = 0,
                 covariate_size: int = 0,
                 method: Literal['flow','vae'] = 'vae',
                 transforms: int = 1,
                 z_dim: int = 30,
                 z_dist: Literal['normal','studentt','laplacian','cauchy','gumbel'] = 'studentt',
                 loss_func: Literal['negbinomial','poisson','multinomial','bernoulli'] = 'multinomial',
                 dispersion: float = 10.0,
                 use_zeroinflate: bool = True,
                 hidden_layers: list = [500],
                 hidden_layer_activation: Literal['relu','softplus','leakyrelu','linear'] = 'relu',
                 flow_hidden_layers = [256],
                 nn_dropout: float = 0.1,
                 post_layer_fct: list = ['layernorm'],
                 post_act_fct: list = None,
                 config_enum: str = 'parallel',
                 use_cuda: bool = True,
                 seed: int = 42,
                 dtype = torch.float32, # type: ignore
                 ):
        super().__init__()
        
        if method == 'flow':
            self.engine = SURENF(input_dim=input_dim,
                                 codebook_size=codebook_size,
                                 condition_size=condition_size,
                                 covariate_size=covariate_size,
                                 transforms=transforms,
                                 z_dim=z_dim,
                                 z_dist=z_dist,
                                 loss_func=loss_func,
                                 dispersion=dispersion,
                                 use_zeroinflate=use_zeroinflate,
                                 hidden_layers=hidden_layers,
                                 hidden_layer_activation=hidden_layer_activation,
                                 flow_hidden_layers=flow_hidden_layers,
                                 nn_dropout=nn_dropout,
                                 post_act_fct=post_act_fct,
                                 post_layer_fct=post_layer_fct,
                                 config_enum=config_enum,
                                 use_cuda=use_cuda,
                                 seed=seed,
                                 dtype=dtype)
        elif method == 'vae':
            self.engine = SUREVAE(input_dim=input_dim,
                                  codebook_size=codebook_size,
                                  condition_size=condition_size,
                                  covariate_size=covariate_size,
                                  z_dim=z_dim,
                                  z_dist=z_dist,
                                  loss_func=loss_func,
                                  dispersion=dispersion,
                                  use_zeroinflate=use_zeroinflate,
                                  hidden_layers=hidden_layers,
                                  hidden_layer_activation=hidden_layer_activation,
                                  nn_dropout=nn_dropout,
                                  post_layer_fct=post_layer_fct,
                                  post_act_fct=post_act_fct,
                                  config_enum=config_enum,
                                  use_cuda=use_cuda,
                                  seed=seed,
                                  dtype=dtype)

    def get_device(self):
        return self.engine.get_device()

    def get_codebook(self):
        """
        Return the mean part of metacell codebook
        """
        return self.engine.get_codebook()

    def get_cell_embedding(self, 
                             xs, 
                             batch_size: int = 1024,
                             show_progress: bool = True):
        """
        Return cells' latent representations

        Parameters
        ----------
        xs
            Single-cell expression matrix. It should be a Numpy array or a Pytorch Tensor.
        batch_size
            Size of batch processing.
        show_progress
            Verbose on or off
        """
        return self.engine.get_cell_embedding(xs=xs, batch_size=batch_size, show_progress=show_progress)
    
    def code(self, xs, batch_size=1024, show_progress=True):
        return self.engine.code(xs=xs, batch_size=batch_size, show_progress=show_progress)
    
    def soft_assignments(self, xs, batch_size=1024, show_progress=True):
        """
        Map cells to metacells and return the probabilistic values of metacell assignments
        """
        return self.engine.soft_assignments(xs=xs, batch_size=batch_size, show_progress=show_progress)
    
    def hard_assignments(self, xs, batch_size=1024, show_progress=True):
        """
        Map cells to metacells and return the assigned metacell identities.
        """
        return self.engine.hard_assignments(xs=xs, batch_size=batch_size, show_progress=show_progress)
    
    def get_condition_effects(self, xs, cs, batch_size=1024, show_progress=True):
        return self.engine.get_condition_effects(xs, cs, batch_size=batch_size, show_progress=show_progress)
    
    def predict_cluster(self, xs, batch_size=1024, show_progress=True):
        return self.engine.predict_cluster(xs, batch_size=batch_size, show_progress=show_progress)
    
    def predict(self, xs, cs, batch_size=1024, show_progress=True):
        """
        Generate gene expression prediction from given cell data and covariates.
        This function can be used for simulating cells' transcription profiles at new conditions.
        
        :param self: SURE model
        :param xs: Cell data at the source condition
        :param cs: Covariates specifying the target condition for generation
        :param batch_size: Data size per batch
        :param show_progress: Toggle on or off message output
        """
        return self.engine.predict(xs, cs, batch_size, show_progress)
    
    def preprocess(self, xs, threshold=0):
        return self.engine.preprocess(xs=xs, threshold=threshold) 
    
    def fit(self, xs, 
            cs = None, 
            fs = None,
            num_epochs: int = 100, 
            learning_rate: float = 0.0001, 
            use_mask: bool = False,
            mask_ratio: float = 0.15,
            batch_size: int = 1000, 
            algo: Literal['adam','rmsprop','adamw'] = 'adam', 
            beta_1: float = 0.9, 
            weight_decay: float = 0.005, 
            decay_rate: float = 0.9,
            config_enum: str = 'parallel',
            threshold: int = 0,
            use_jax: bool = False,
            show_progress: bool = True,
            # Early stopping 相关参数
            patience: int = 10,
            min_delta: float = 1e-4,
            restore_best_weights: bool = True,
            monitor: str = 'loss'):
        """
        Train the SURE model.

        Parameters
        ----------
        xs
            Single-cell experssion matrix. It should be a Numpy array or a Pytorch Tensor. Rows are cells and columns are features.
        us
            cell-level factor matrix. 
        num_epochs
            Number of training epochs.
        learning_rate
            Parameter for training.
        batch_size
            Size of batch processing.
        algo
            Optimization algorithm.
        beta_1
            Parameter for optimization.
        weight_decay
            Parameter for optimization.
        decay_rate 
            Parameter for optimization.
        use_jax
            If toggled on, Jax will be used for speeding up. CAUTION: This will raise errors because of unknown reasons when it is called in
            the Python script or Jupyter notebook. It is OK if it is used when runing SURE in the shell command.
        """
        self.engine.fit(xs=xs, cs=cs, fs=fs, num_epochs=num_epochs, learning_rate=learning_rate, use_mask=use_mask, mask_ratio=mask_ratio, batch_size=batch_size, algo=algo,
                        beta_1=beta_1, weight_decay=weight_decay, decay_rate=decay_rate, config_enum=config_enum, threshold=threshold,
                        use_jax=use_jax, show_progress=show_progress, patience=patience, min_delta=min_delta, restore_best_weights=restore_best_weights,
                        monitor=monitor)
        
    def build_network(self, xs, 
                      k_data_to_data:int=5,
                      k_data_to_centers:int=3,
                      k_centers_to_data:int=30, 
                      k_centers_to_centers:int=7, 
                      sample_size:int=3000,
                      temperature:float=1.0,
                      builder:Literal['bipartite','knn']='bipartite',
                      weight_method:Literal['count','proportion', 'weighted']='weighted',
                      metric:Literal['euclidean','correlation','cosine','mahalanobis']='cosine',
                      min_edge_weight:float=0.01,
                      normalize_weights:bool=True,
                      show_progress:bool=False):
        
        n_sample = xs.shape[0]
        # metacell attributes
        zs = self.engine.get_cell_embedding(xs, show_progress=show_progress)
        
        pca_model = PCA().fit(self.engine.codebook_loc)
        zs = pca_model.transform(zs)
        cb = pca_model.transform(self.engine.codebook_loc)
        
        idx = np.arange(n_sample)
        if n_sample>sample_size:
            idx = np.random.choice(idx, size=sample_size)
            
        # create network
        if builder=='bipartite':
            network_builder = ClusterNetworkBuilder(weight_method=weight_method, 
                                                    k_data_to_centers=k_data_to_centers, k_centers_to_data=k_centers_to_data,
                                                    min_edge_weight=min_edge_weight, normalize_weights=normalize_weights)
            G = network_builder.build_network(zs[idx], cb, metric=metric)
        else:
            network_builder = KnnClusterNetworkBuilder(k_data=k_data_to_data, k_centers=k_centers_to_data,
                                                       temperature=temperature, metric=metric)
            G = network_builder.build_network(zs[idx], cb)
        
        y = self.predict_cluster(xs, show_progress=show_progress)
        Y = np.zeros([n_sample, self.engine.code_dim])
        Y[np.arange(n_sample),y] = 1 
        nY = np.sum(Y, axis=0)
        
        for i in np.arange(self.engine.code_dim):
            G.nodes[i]['cluster_size'] = nY[i]
            G.nodes[i]['cluster_weight'] = nY[i]/n_sample
            G.nodes[i]['weight'] = nY[i]/n_sample
        
        G = keep_top_k_edges_per_node(G, k_centers_to_centers, inplace=True)
        
        # assign data points to edges
        edge_allocator = EdgeAllocation()
        edge_weights = edge_allocator.fit_transform(zs[idx], cb, G, metric=metric)
        for edge in edge_weights.keys():
            i,j = edge 
            G[i][j]['weight'] = edge_weights[edge]
            
        nodes_data = [
            {'node_id': node, **attrs} 
            for node, attrs in G.nodes(data=True)
        ]
        
        edges_data = [
            {'source': u, 'target': v, **attrs}
            for u, v, attrs in G.edges(data=True)
        ]
        
        return G,pd.DataFrame(nodes_data),pd.DataFrame(edges_data)
    
    def display_network(self, G,
                        node_size_attr='weight',
                        node_size_scale=5000, edge_size_scale=10, 
                        use_plotly=False,
                        figsize=(12,10)):       
        # 元细胞坐标点
        pos_array = PCA().fit_transform(self.engine.codebook_loc)
        
        if not use_plotly:
            visualize_network(G, pos_array, node_size_attr=node_size_attr, node_size_scale=node_size_scale, edge_size_scale=edge_size_scale, figsize=figsize)
        else:
            visualize_network_interactive(G, pos_array, node_size_scale=node_size_scale, edge_size_scale=edge_size_scale, figsize=figsize*100)

    @classmethod
    def save_model(cls, model, file_path, compression=False):
        """Save the model to the specified file path."""
        file_path = os.path.abspath(file_path)

        model.eval()
        if compression:
            with gzip.open(file_path, 'wb') as pickle_file:
                pickle.dump(model, pickle_file)
        else:
            with open(file_path, 'wb') as pickle_file:
                pickle.dump(model, pickle_file)

        print(f'Model saved to {file_path}')

    @classmethod
    def load_model(cls, file_path):
        """Load the model from the specified file path and return an instance."""
        print(f'Model loaded from {file_path}')

        file_path = os.path.abspath(file_path)
        if file_path.endswith('gz'):
            with gzip.open(file_path, 'rb') as pickle_file:
                model = pickle.load(pickle_file)
        else:
            with open(file_path, 'rb') as pickle_file:
                model = pickle.load(pickle_file)
        
        return model

        



# This provides the implementation of negative binomial distributions used in SURE.
# The code is modified from NegativeBinomial of scvi-tools.

from __future__ import annotations

import warnings

import torch
import torch.nn.functional as F
import torch.distributions as dist
from torch.distributions import Distribution, Gamma, constraints
from torch.distributions import Poisson as PoissonTorch
from torch.distributions.utils import (
    broadcast_all,
    lazy_property,
    logits_to_probs,
    probs_to_logits,
)




def torch_lgamma_mps(x: torch.Tensor) -> torch.Tensor:
    """Used in mac Mx devices while broadcasting a tensor

    Parameters
    ----------
    x
        Data

    Returns
    -------
    lgamma tensor that perform on a copied version of the tensor
    """
    return torch.lgamma(x.contiguous())


def log_zinb_positive(
    x: torch.Tensor,
    mu: torch.Tensor,
    theta: torch.Tensor,
    pi: torch.Tensor,
    eps: float = 1e-8,
    log_fn: callable = torch.log,
    lgamma_fn: callable = torch.lgamma,
) -> torch.Tensor:
    """Log likelihood (scalar) of a minibatch according to a zinb model.

    Parameters
    ----------
    x
        Data
    mu
        mean of the negative binomial (has to be positive support) (shape: minibatch x vars)
    theta
        inverse dispersion parameter (has to be positive support) (shape: minibatch x vars)
    pi
        logit of the dropout parameter (real support) (shape: minibatch x vars)
    eps
        numerical stability constant
    log_fn
        log function
    lgamma_fn
        log gamma function

    Notes
    -----
    We parametrize the bernoulli using the logits, hence the softplus functions appearing.
    """
    log = log_fn
    lgamma = lgamma_fn
    # theta is the dispersion rate. If .ndimension() == 1, it is shared for all cells (regardless
    # of batch or labels)
    if theta.ndimension() == 1:
        theta = theta.view(1, theta.size(0))  # In this case, we reshape theta for broadcasting

    # Uses log(sigmoid(x)) = -softplus(-x)
    softplus_pi = F.softplus(-pi)
    log_theta_eps = log(theta + eps)
    log_theta_mu_eps = log(theta + mu + eps)
    pi_theta_log = -pi + theta * (log_theta_eps - log_theta_mu_eps)

    case_zero = F.softplus(pi_theta_log) - softplus_pi
    mul_case_zero = torch.mul((x < eps).type(torch.float32), case_zero)

    case_non_zero = (
        -softplus_pi
        + pi_theta_log
        + x * (log(mu + eps) - log_theta_mu_eps)
        + lgamma(x + theta)
        - lgamma(theta)
        - lgamma(x + 1)
    )
    mul_case_non_zero = torch.mul((x > eps).type(torch.float32), case_non_zero)

    res = mul_case_zero + mul_case_non_zero

    return res


def log_nb_positive(
    x: torch.Tensor,
    mu: torch.Tensor,
    theta: torch.Tensor,
    eps: float = 1e-8,
    log_fn: callable = torch.log,
    lgamma_fn: callable = torch.lgamma,
) -> torch.Tensor:
    """Log likelihood (scalar) of a minibatch according to a nb model.

    Parameters
    ----------
    x
        data
    mu
        mean of the negative binomial (has to be positive support) (shape: minibatch x vars)
    theta
        inverse dispersion parameter (has to be positive support) (shape: minibatch x vars)
    eps
        numerical stability constant
    log_fn
        log function
    lgamma_fn
        log gamma function
    """
    log = log_fn
    lgamma = lgamma_fn
    log_theta_mu_eps = log(theta + mu + eps)
    res = (
        theta * (log(theta + eps) - log_theta_mu_eps)
        + x * (log(mu + eps) - log_theta_mu_eps)
        + lgamma(x + theta)
        - lgamma(theta)
        - lgamma(x + 1)
    )

    return res


def _convert_mean_disp_to_counts_logits(
    mu: torch.Tensor,
    theta: torch.Tensor,
    eps: float = 1e-6,
) -> tuple[torch.Tensor, torch.Tensor]:
    r"""NB parameterizations conversion.

    Parameters
    ----------
    mu
        mean of the NB distribution.
    theta
        inverse overdispersion.
    eps
        constant used for numerical log stability. (Default value = 1e-6)

    Returns
    -------
    type
        the number of failures until the experiment is stopped
        and the success probability.
    """
    if not (mu is None) == (theta is None):
        raise ValueError(
            "If using the mu/theta NB parameterization, both parameters must be specified"
        )
    logits = (mu + eps).log() - (theta + eps).log()
    total_count = theta
    return total_count, logits


def _convert_counts_logits_to_mean_disp(
    total_count: torch.Tensor, logits: torch.Tensor
) -> tuple[torch.Tensor, torch.Tensor]:
    """NB parameterizations conversion.

    Parameters
    ----------
    total_count
        Number of failures until the experiment is stopped.
    logits
        success logits.

    Returns
    -------
    type
        the mean and inverse overdispersion of the NB distribution.

    """
    theta = total_count
    mu = logits.exp() * theta
    return mu, theta


def _gamma(theta: torch.Tensor, mu: torch.Tensor, on_mps: bool = False) -> Gamma:
    concentration = theta
    rate = theta / mu
    # Important remark: Gamma is parametrized by the rate = 1/scale!
    gamma_d = (
        Gamma(concentration=concentration.to("cpu"), rate=rate.to("cpu"))
        if on_mps  # TODO: NEED TORCH MPS FIX for 'aten::_standard_gamma'
        else Gamma(concentration=concentration, rate=rate)
    )
    return gamma_d


class NegativeBinomial(Distribution):
    r"""Negative binomial distribution.

    Parameters
    ----------
    mu
        Mean of the distribution.
    theta
        Inverse dispersion.
    scale
        Normalized mean expression of the distribution.
    validate_args
        Raise ValueError if arguments do not match constraints
    """

    arg_constraints = {
        "mu": constraints.greater_than_eq(0),
        "theta": constraints.greater_than_eq(0),
        "scale": constraints.greater_than_eq(0),
    }
    #support = constraints.nonnegative_integer
    support = constraints.nonnegative

    def __init__(
        self,
        mu: torch.Tensor | None = None,
        theta: torch.Tensor | None = None,
        scale: torch.Tensor | None = None,
        validate_args: bool = False,
    ):
        self.on_mps = (
            mu.device.type == "mps"
        )  # TODO: This is used until torch will solve the MPS issues
        self._eps = 1e-8

        mu, theta = broadcast_all(mu, theta)
        if self.on_mps:
            mu, theta = mu.contiguous(), theta.contiguous()
        self.mu = mu
        self.theta = theta
        self.scale = scale
        
        batch_shape = self.mu.shape
        event_shape = torch.Size()
        super().__init__(batch_shape, event_shape, validate_args=validate_args)

    @property
    def mean(self) -> torch.Tensor:
        return self.mu

    def get_normalized(self, key) -> torch.Tensor:
        if key == "mu":
            return self.mu
        elif key == "scale":
            return self.scale
        else:
            raise ValueError(f"normalized key {key} not recognized")

    @property
    def variance(self) -> torch.Tensor:
        return self.mean + (self.mean**2) / self.theta

    @torch.inference_mode()
    def sample(
        self,
        sample_shape: torch.Size | tuple | None = None,
    ) -> torch.Tensor:
        """Sample from the distribution."""
        sample_shape = sample_shape or torch.Size()
        gamma_d = self._gamma()  # TODO: TORCH MPS FIX - DONE ON CPU CURRENTLY
        p_means = gamma_d.sample(sample_shape)

        # Clamping as distributions objects can have buggy behaviors when
        # their parameters are too high
        l_train = torch.clamp(p_means, max=1e8)
        counts = (
            PoissonTorch(l_train).sample().to("mps")
            if self.on_mps  # TODO: NEED TORCH MPS FIX for 'aten::poisson'
            else PoissonTorch(l_train).sample()
        )  # Shape : (n_samples, n_cells_batch, n_vars)
        return counts

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        if self._validate_args:
            try:
                self._validate_sample(value)
            except ValueError:
                warnings.warn(
                    "The value argument must be within the support of the distribution",
                    UserWarning,
                )

        lgamma_fn = torch_lgamma_mps if self.on_mps else torch.lgamma  # TODO: TORCH MPS FIX
        return log_nb_positive(
            value, mu=self.mu, theta=self.theta, eps=self._eps, lgamma_fn=lgamma_fn
        )

    def _gamma(self) -> Gamma:
        return _gamma(self.theta, self.mu, self.on_mps)

    def to_event(self, n):
        if n == 0:
            return self
        return dist.Independent(self, n)
    
    def __repr__(self) -> str:
        param_names = [k for k, _ in self.arg_constraints.items() if k in self.__dict__]
        args_string = ", ".join(
            [
                f"{p}: "
                f"{self.__dict__[p] if self.__dict__[p].numel() == 1 else self.__dict__[p].size()}"
                for p in param_names
                if self.__dict__[p] is not None
            ]
        )
        return self.__class__.__name__ + "(" + args_string + ")"



class ZeroInflatedNegativeBinomial(NegativeBinomial):
    r"""Zero-inflated negative binomial distribution.

    Parameters
    ----------
    mu
        Mean of the distribution.
    theta
        Inverse dispersion.
    zi_logits
        Logits scale of zero inflation probability.
    scale
        Normalized mean expression of the distribution.
    validate_args
        Raise ValueError if arguments do not match constraints
    """

    arg_constraints = {
        "mu": constraints.greater_than_eq(0),
        "theta": constraints.greater_than_eq(0),
        "zi_logits": constraints.real,
        "scale": constraints.greater_than_eq(0),
    }
    #support = constraints.nonnegative_integer
    support = constraints.nonnegative

    def __init__(
        self,
        mu: torch.Tensor | None = None,
        theta: torch.Tensor | None = None,
        zi_logits: torch.Tensor | None = None,
        scale: torch.Tensor | None = None,
        validate_args: bool = False,
    ):
        super().__init__(
            mu=mu,
            theta=theta,
            scale=scale,
            validate_args=validate_args,
        )
        self.zi_logits, self.mu, self.theta = broadcast_all(zi_logits, self.mu, self.theta)

    @property
    def mean(self) -> torch.Tensor:
        pi = self.zi_probs
        return (1 - pi) * self.mu

    @property
    def variance(self) -> None:
        pi = self.zi_probs
        return (1 - pi) * self.mu * (self.mu + self.theta + pi * self.mu * self.theta) / self.theta

    @lazy_property
    def zi_logits(self) -> torch.Tensor:
        """ZI logits."""
        return probs_to_logits(self.zi_probs, is_binary=True)

    @lazy_property
    def zi_probs(self) -> torch.Tensor:
        return logits_to_probs(self.zi_logits, is_binary=True)

    @torch.inference_mode()
    def sample(
        self,
        sample_shape: torch.Size | tuple | None = None,
    ) -> torch.Tensor:
        """Sample from the distribution."""
        sample_shape = sample_shape or torch.Size()
        samp = super().sample(sample_shape=sample_shape)
        is_zero = torch.rand_like(samp) <= self.zi_probs
        samp_ = torch.where(is_zero, torch.zeros_like(samp), samp)
        return samp_

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """Log probability."""
        try:
            self._validate_sample(value)
        except ValueError:
            warnings.warn(
                "The value argument must be within the support of the distribution",
                UserWarning,
            )
        lgamma_fn = torch_lgamma_mps if self.on_mps else torch.lgamma  # TODO: TORCH MPS FIX
        return log_zinb_positive(
            value, self.mu, self.theta, self.zi_logits, eps=1e-08, lgamma_fn=lgamma_fn
        )
        
    def to_event(self, n):
        if n == 0:
            return self
        return dist.Independent(self, n)


from typing import Union, Optional
from warnings import warn

import numpy as np
from scipy.sparse import csr_matrix, dia_matrix, issparse
from sklearn.preprocessing import MinMaxScaler

from anndata import AnnData

import scanpy as sc
from scanpy._utils import view_to_actual


# Computational methods for preprocessing


def tfidf(
    data: AnnData,
    log_tf: bool = True,
    log_idf: bool = True,
    log_tfidf: bool = False,
    scale_factor: Union[int, float] = 1e4,
    inplace: bool = True,
    copy: bool = False,
    from_layer: Optional[str] = None,
    to_layer: Optional[str] = None,
):
    """
    Transform peak counts with TF-IDF (Term Frequency - Inverse Document Frequency).

    TF: peak counts are normalised by total number of counts per cell
    DF: total number of counts for each peak
    IDF: number of cells divided by DF

    By default, log(TF) * log(IDF) is returned.

    Parameters
    ----------
    data
            AnnData object with peak counts.
    log_idf
            Log-transform IDF term (True by default).
    log_tf
            Log-transform TF term (True by default).
    log_tfidf
            Log-transform TF*IDF term (False by default).
            Can only be used when log_tf and log_idf are False.
    scale_factor
            Scale factor to multiply the TF-IDF matrix by (1e4 by default).
    inplace
            If to modify counts in the AnnData object (True by default).
    copy
            If to return a copy of the AnnData object or the 'atac' modality (False by default).
            Not compatible with inplace=False.
    from_layer
            Layer to use counts (AnnData.layers[from_layer])
            instead of AnnData.X used by default.
    to_layer
            Layer to save transformed counts to (AnnData.layers[to_layer])
            instead of AnnData.X used by default.
            Not compatible with inplace=False.
    """
    if isinstance(data, AnnData):
        adata = data
    else:
        raise TypeError("Expected AnnData object")

    if log_tfidf and (log_tf or log_idf):
        raise AttributeError(
            "When returning log(TF*IDF), \
            applying neither log(TF) nor log(IDF) is possible."
        )

    if copy and not inplace:
        raise ValueError("`copy=True` cannot be used with `inplace=False`.")

    if to_layer is not None and not inplace:
        raise ValueError(f"`to_layer='{str(to_layer)}'` cannot be used with `inplace=False`.")

    if copy:
        adata = adata.copy()

    view_to_actual(adata)

    counts = adata.X if from_layer is None else adata.layers[from_layer]

    # Check before the computation
    if to_layer is not None and to_layer in adata.layers:
        warn(f"Existing layer '{str(to_layer)}' will be overwritten")

    if issparse(counts):
        n_peaks = np.asarray(counts.sum(axis=1)).reshape(-1)
        n_peaks = dia_matrix((1.0 / n_peaks, 0), shape=(n_peaks.size, n_peaks.size))
        # This prevents making TF dense
        tf = np.dot(n_peaks, counts)
    else:
        n_peaks = np.asarray(counts.sum(axis=1)).reshape(-1, 1)
        tf = counts / n_peaks

    if scale_factor is not None and scale_factor != 0 and scale_factor != 1:
        tf = tf * scale_factor
    if log_tf:
        tf = np.log1p(tf)

    idf = np.asarray(adata.shape[0] / counts.sum(axis=0)).reshape(-1)
    if log_idf:
        idf = np.log1p(idf)

    if issparse(tf):
        idf = dia_matrix((idf, 0), shape=(idf.size, idf.size))
        tf_idf = np.dot(tf, idf)
    else:
        tf_idf = np.dot(csr_matrix(tf), csr_matrix(np.diag(idf)))

    if log_tfidf:
        tf_idf = np.log1p(tf_idf)

    res = np.nan_to_num(tf_idf, nan=0.0)
    if not inplace:
        return res

    if to_layer is not None:
        adata.layers[to_layer] = res
    else:
        adata.X = res

    if copy:
        return adata


def binarize(data: AnnData, threshold: np.float32=0):
    """
    Transform peak counts to the binary matrix (all the non-zero values become 1).

    Parameters
    ----------
    data
            AnnData object with peak counts.
    """
    if isinstance(data, AnnData):
        adata = data
    else:
        raise TypeError("Expected AnnData object")

    if issparse(adata.X):
        # Sparse matrix
        adata.X.data[adata.X.data > threshold] = 1
    else:
        adata.X[adata.X > threshold] = 1





import pyro
import pyro.distributions as dist
from pyro.optim import ExponentialLR
from pyro.infer import SVI, JitTraceEnum_ELBO, TraceEnum_ELBO, config_enumerate

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.distributions.utils import logits_to_probs, probs_to_logits, clamp_probs
from torch.distributions import constraints
from torch.distributions.transforms import SoftmaxTransform

from .utils.custom_mlp import MLP, Exp, ZeroBiasMLP
from .utils.utils import CustomDataset, CustomDataset2, CustomDataset4, tensor_to_numpy, convert_to_tensor

from .dist.negbinomial import NegativeBinomial as MyNB
from .dist.negbinomial import ZeroInflatedNegativeBinomial as MyZINB

import zuko 
from pyro.contrib.zuko import ZukoToPyro

import os
import argparse
import random
import numpy as np
import datatable as dt
from tqdm import tqdm
from scipy import sparse

import scanpy as sc
from .atac import binarize

from typing import Literal

import warnings
warnings.filterwarnings("ignore")

import dill as pickle
import gzip
from packaging.version import Version
torch_version = torch.__version__


def set_random_seed(seed):
    # Set seed for PyTorch
    torch.manual_seed(seed)
    
    # If using CUDA, set the seed for CUDA
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)  # For multi-GPU setups.
    
    # Set seed for NumPy
    np.random.seed(seed)
    
    # Set seed for Python's random module
    random.seed(seed)

    # Set seed for Pyro
    pyro.set_rng_seed(seed)

class SUREVanilla(nn.Module):
    """SUccinct REpresentation of single-omics cells without codebook

    Parameters
    ----------
    inpute_size
        Number of features (e.g., genes, peaks, proteins, etc.) per cell.
    covariate_size
        Number of cell-level factors. 
    transforms
        Number of neural spline flows
    z_dim
        Dimensionality of latent states and metacells. 
    hidden_layers
        A list gives the numbers of neurons for each hidden layer.
    flow_hidden_layers
        A list gives the numbers of neurons for each hidden layer in neural spline flows.
    loss_func
        The likelihood model for single-cell data generation. 
        
        One of the following: 
        * ``'negbinomial'`` -  negative binomial distribution (default)
        * ``'poisson'`` - poisson distribution
        * ``'multinomial'`` - multinomial distribution
    use_cuda
        A boolean option for switching on cuda device. 

    Examples
    --------
    >>>
    >>>
    >>>

    """
    def __init__(self,
                 input_dim: int,
                 covariate_size: int = 0,
                 transforms: int = 1,
                 z_dim: int = 50,
                 z_dist: Literal['normal','studentt','laplacian','cauchy','gumbel'] = 'studentt',
                 loss_func: Literal['negbinomial','poisson','multinomial','bernoulli'] = 'negbinomial',
                 dispersion: float = 10.0,
                 use_zeroinflate: bool = True,
                 hidden_layers: list = [500],
                 hidden_layer_activation: Literal['relu','softplus','leakyrelu','linear'] = 'relu',
                 flow_hidden_layers = [128,128,128],
                 nn_dropout: float = 0.1,
                 post_layer_fct: list = ['layernorm'],
                 post_act_fct: list = None,
                 config_enum: str = 'parallel',
                 use_cuda: bool = True,
                 seed: int = 42,
                 dtype = torch.float32, # type: ignore
                 ):
        super().__init__()

        self.input_dim = input_dim
        self.covariate_size = covariate_size
        self.dispersion = dispersion
        self.latent_dim = z_dim
        self.latent_dist = z_dist
        self.hidden_layers = hidden_layers
        self.decoder_hidden_layers = hidden_layers[::-1]
        self.allow_broadcast = config_enum == 'parallel'
        self.use_cuda = use_cuda
        self.loss_func = loss_func
        self.options = None
        self.dtype = dtype
        self.use_zeroinflate=use_zeroinflate
        self.nn_dropout = nn_dropout
        self.post_layer_fct = post_layer_fct
        self.post_act_fct = post_act_fct
        self.hidden_layer_activation = hidden_layer_activation
        self.transforms = transforms
        self.flow_hidden_layers = flow_hidden_layers
        
        set_random_seed(seed)
        self.setup_networks()

    def setup_networks(self):
        latent_dim = self.latent_dim
        hidden_sizes = self.hidden_layers

        nn_layer_norm, nn_batch_norm, nn_layer_dropout = False, False, False
        na_layer_norm, na_batch_norm, na_layer_dropout = False, False, False

        if self.post_layer_fct is not None:
            nn_layer_norm=True if ('layernorm' in self.post_layer_fct) or ('layer_norm' in self.post_layer_fct) else False
            nn_batch_norm=True if ('batchnorm' in self.post_layer_fct) or ('batch_norm' in self.post_layer_fct) else False
            nn_layer_dropout=True if 'dropout' in self.post_layer_fct else False

        if self.post_act_fct is not None:
            na_layer_norm=True if ('layernorm' in self.post_act_fct) or ('layer_norm' in self.post_act_fct) else False
            na_batch_norm=True if ('batchnorm' in self.post_act_fct) or ('batch_norm' in self.post_act_fct) else False
            na_layer_dropout=True if 'dropout' in self.post_act_fct else False

        if nn_layer_norm and nn_batch_norm and nn_layer_dropout:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout),nn.BatchNorm1d(layer.module.out_features), nn.LayerNorm(layer.module.out_features))
        elif nn_layer_norm and nn_layer_dropout:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout), nn.LayerNorm(layer.module.out_features))
        elif nn_batch_norm and nn_layer_dropout:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout), nn.BatchNorm1d(layer.module.out_features))
        elif nn_layer_norm and nn_batch_norm:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.BatchNorm1d(layer.module.out_features), nn.LayerNorm(layer.module.out_features))
        elif nn_layer_norm:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.LayerNorm(layer.module.out_features)
        elif nn_batch_norm:
            post_layer_fct = lambda layer_ix, total_layers, layer:nn.BatchNorm1d(layer.module.out_features)
        elif nn_layer_dropout:
            post_layer_fct = lambda layer_ix, total_layers, layer: nn.Dropout(self.nn_dropout)
        else:
            post_layer_fct = lambda layer_ix, total_layers, layer: None

        if na_layer_norm and na_batch_norm and na_layer_dropout:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout),nn.BatchNorm1d(layer.module.out_features), nn.LayerNorm(layer.module.out_features))
        elif na_layer_norm and na_layer_dropout:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout), nn.LayerNorm(layer.module.out_features))
        elif na_batch_norm and na_layer_dropout:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.Dropout(self.nn_dropout), nn.BatchNorm1d(layer.module.out_features))
        elif na_layer_norm and na_batch_norm:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Sequential(nn.BatchNorm1d(layer.module.out_features), nn.LayerNorm(layer.module.out_features))
        elif na_layer_norm:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.LayerNorm(layer.module.out_features)
        elif na_batch_norm:
            post_act_fct = lambda layer_ix, total_layers, layer:nn.BatchNorm1d(layer.module.out_features)
        elif na_layer_dropout:
            post_act_fct = lambda layer_ix, total_layers, layer: nn.Dropout(self.nn_dropout)
        else:
            post_act_fct = lambda layer_ix, total_layers, layer: None

        if self.hidden_layer_activation == 'relu':
            activate_fct = nn.ReLU
        elif self.hidden_layer_activation == 'softplus':
            activate_fct = nn.Softplus
        elif self.hidden_layer_activation == 'leakyrelu':
            activate_fct = nn.LeakyReLU
        elif self.hidden_layer_activation == 'linear':
            activate_fct = nn.Identity

        if self.covariate_size>0:
            self.covariate_effect = ZeroBiasMLP(
                [self.latent_dim + self.covariate_size] + self.decoder_hidden_layers + [self.latent_dim],
                activation=activate_fct,
                output_activation=None,
                post_layer_fct=post_layer_fct,
                post_act_fct=post_act_fct,
                allow_broadcast=self.allow_broadcast,
                use_cuda=self.use_cuda,
            )
            
        self.decoder_log_mu = MLP(
                [self.latent_dim] + self.decoder_hidden_layers + [self.input_dim],
                activation=activate_fct,
                output_activation=None,
                post_layer_fct=post_layer_fct,
                post_act_fct=post_act_fct,
                allow_broadcast=self.allow_broadcast,
                use_cuda=self.use_cuda,
            )
        
        self.encoder_zn = MLP(
            [self.input_dim] + hidden_sizes + [[latent_dim, latent_dim]],
            activation=activate_fct,
            output_activation=[None, Exp],
            post_layer_fct=post_layer_fct,
            post_act_fct=post_act_fct,
            allow_broadcast=self.allow_broadcast,
            use_cuda=self.use_cuda,
        )
        
        if self.use_cuda:
            self.cuda()

    def get_device(self):
        return next(self.parameters()).device

    def cutoff(self, xs, thresh=None):
        eps = torch.finfo(xs.dtype).eps
        
        if not thresh is None:
            if eps < thresh:
                eps = thresh

        xs = xs.clamp(min=eps)

        if torch.any(torch.isnan(xs)):
            xs[torch.isnan(xs)] = eps

        return xs

    def softmax(self, xs):
        #xs = SoftmaxTransform()(xs)
        xs = dist.Multinomial(total_count=1, logits=xs).mean
        return xs
    
    def sigmoid(self, xs):
        #sigm_enc = nn.Sigmoid()
        #xs = sigm_enc(xs)
        #xs = clamp_probs(xs)
        xs = dist.Bernoulli(logits=xs).mean
        return xs

    def softmax_logit(self, xs):
        eps = torch.finfo(xs.dtype).eps
        xs = self.softmax(xs)
        xs = torch.logit(xs, eps=eps)
        return xs

    def logit(self, xs):
        eps = torch.finfo(xs.dtype).eps
        xs = torch.logit(xs, eps=eps)
        return xs

    def dirimulti_param(self, xs):
        xs = self.dirimulti_mass * self.sigmoid(xs)
        return xs

    def multi_param(self, xs):
        xs = self.softmax(xs)
        return xs

    def model(self, xs, cs=None):
        pyro.module('sure', self)

        eps = torch.finfo(xs.dtype).eps
        batch_size = xs.size(0)
        self.options = dict(dtype=xs.dtype, device=xs.device)
        
        if self.loss_func=='negbinomial':
            dispersion = pyro.param("dispersion", self.dispersion *
                                     xs.new_ones(self.input_dim), constraint=constraints.positive)
            
        if self.use_zeroinflate:
            gate_logits = pyro.param("dropout_rate", xs.new_zeros(self.input_dim))

        with pyro.plate('data'):
            zn_loc = torch.zeros(batch_size, self.latent_dim, **self.options)
            zn_scale = torch.ones(batch_size, self.latent_dim, **self.options)
            if self.latent_dist == 'studentt':
                zn_dof = torch.ones(batch_size, self.latent_dim, **self.options)
                zns = pyro.sample('zn', dist.StudentT(df=zn_dof, loc=zn_loc, scale=zn_scale).to_event(1))
            elif self.latent_dist == 'laplacian':
                zns = pyro.sample('zn', dist.Laplace(zn_loc, zn_scale).to_event(1))
            elif self.latent_dist == 'cauchy':
                zns = pyro.sample('zn', dist.Cauchy(zn_loc, zn_scale).to_event(1))
            elif self.latent_dist == 'normal':
                zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))
            elif self.latent_dist == 'gumbel':
                zns = pyro.sample('zn', dist.Gumbel(zn_loc, zn_scale).to_event(1))
                
            if (self.covariate_size>0) and (cs is not None):
                zus = self.covariate_effect([zns,cs])
                zs = zns+zus
            else:
                zs = zns

            log_mu = self.decoder_log_mu(zs)
            if self.loss_func in ['bernoulli']:
                log_theta = log_mu
            elif self.loss_func in ['negbinomial']:
                mu = log_mu.exp()
            else:
                rate = log_mu.exp()
                theta = dist.DirichletMultinomial(total_count=1, concentration=rate).mean
                if self.loss_func == 'poisson':
                    rate = theta * torch.sum(xs, dim=1, keepdim=True)

            if self.loss_func == 'negbinomial':
                if self.use_zeroinflate:
                    pyro.sample("x", MyZINB(mu=mu, theta=dispersion, zi_logits=gate_logits).to_event(1), obs=xs)
                else:
                    pyro.sample("x", MyNB(mu=mu, theta=dispersion).to_event(1), obs=xs)
            elif self.loss_func == 'poisson':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Poisson(rate=rate),gate_logits=gate_logits).to_event(1), obs=xs.round())
                else:
                    pyro.sample('x', dist.Poisson(rate=rate).to_event(1), obs=xs.round())
            elif self.loss_func == 'multinomial':
                pyro.sample('x', dist.Multinomial(total_count=int(1e8), probs=theta), obs=xs)
            elif self.loss_func == 'bernoulli':
                if self.use_zeroinflate:
                    pyro.sample('x', dist.ZeroInflatedDistribution(dist.Bernoulli(logits=log_theta),gate_logits=gate_logits).to_event(1), obs=xs)
                else:
                    pyro.sample('x', dist.Bernoulli(logits=log_theta).to_event(1), obs=xs)

    def guide(self, xs, cs=None):
        with pyro.plate('data'):
            zn_loc, zn_scale = self.encoder_zn(xs)
            zns = pyro.sample('zn', dist.Normal(zn_loc, zn_scale).to_event(1))

    def _get_cell_embedding(self, xs):           
        zns, _ = self.encoder_zn(xs)
        return zns 
    
    def get_cell_embedding(self, 
                             xs, 
                             batch_size: int = 1024,
                             show_progress: bool = True):
        """
        Return cells' latent representations

        Parameters
        ----------
        xs
            Single-cell expression matrix. It should be a Numpy array or a Pytorch Tensor.
        batch_size
            Size of batch processing.
        show_progress
            Verbose on or off
        """
        xs = self.preprocess(xs)
        xs = convert_to_tensor(xs, device='cpu')
        dataset = CustomDataset(xs)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)

        Z = []
        with tqdm(total=len(dataloader), disable=not show_progress, desc='', unit='batch') as pbar:
            for X_batch, _ in dataloader:
                X_batch = X_batch.to(self.get_device())
                zns = self._get_cell_embedding(X_batch)
                Z.append(tensor_to_numpy(zns))
                pbar.update(1)

        Z = np.concatenate(Z)
        return Z
    
    def preprocess(self, xs, threshold=0):
        if self.loss_func == 'bernoulli':
            ad = sc.AnnData(xs)
            binarize(ad, threshold=threshold)
            xs = ad.X.copy()
        else:
            xs = np.round(xs)
            
        if sparse.issparse(xs):
            xs = xs.toarray()
        return xs 
    
    def fit(self, xs, 
            cs = None, 
            num_epochs: int = 100, 
            learning_rate: float = 0.0001, 
            batch_size: int = 256, 
            algo: Literal['adam','rmsprop','adamw'] = 'adam', 
            beta_1: float = 0.9, 
            weight_decay: float = 0.005, 
            decay_rate: float = 0.9,
            config_enum: str = 'parallel',
            threshold: int = 0,
            use_jax: bool = True,
            show_progress: bool = True):
        """
        Train the SURE model.

        Parameters
        ----------
        xs
            Single-cell experssion matrix. It should be a Numpy array or a Pytorch Tensor. Rows are cells and columns are features.
        us
            cell-level factor matrix. 
        num_epochs
            Number of training epochs.
        learning_rate
            Parameter for training.
        batch_size
            Size of batch processing.
        algo
            Optimization algorithm.
        beta_1
            Parameter for optimization.
        weight_decay
            Parameter for optimization.
        decay_rate 
            Parameter for optimization.
        use_jax
            If toggled on, Jax will be used for speeding up. CAUTION: This will raise errors because of unknown reasons when it is called in
            the Python script or Jupyter notebook. It is OK if it is used when runing SURE in the shell command.
        """
        xs = self.preprocess(xs, threshold=threshold)
        xs = convert_to_tensor(xs, dtype=self.dtype, device='cpu')
        if cs is not None:
            cs = convert_to_tensor(cs, dtype=self.dtype, device='cpu')

        if cs is None:
            dataset = CustomDataset2(xs, xs)
        else:
            dataset = CustomDataset2(xs, cs)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

        # setup the optimizer
        optim_params = {'lr': learning_rate, 'betas': (beta_1, 0.999), 'weight_decay': weight_decay}

        if algo.lower()=='rmsprop':
            optimizer = torch.optim.RMSprop
        elif algo.lower()=='adam':
            optimizer = torch.optim.Adam
        elif algo.lower() == 'adamw':
            optimizer = torch.optim.AdamW
        else:
            raise ValueError("An optimization algorithm must be specified.")
        scheduler = ExponentialLR({'optimizer': optimizer, 'optim_args': optim_params, 'gamma': decay_rate})

        pyro.clear_param_store()

        # set up the loss(es) for inference, wrapping the guide in config_enumerate builds the loss as a sum
        # by enumerating each class label form the sampled discrete categorical distribution in the model
        Elbo = JitTraceEnum_ELBO if use_jax else TraceEnum_ELBO
        elbo = Elbo(max_plate_nesting=1, strict_enumeration_warning=False)
        guide = config_enumerate(self.guide, config_enum, expand=True)
        loss_basic = SVI(self.model, guide, scheduler, loss=elbo)

        # build a list of all losses considered
        losses = [loss_basic]
        num_losses = len(losses)

        with tqdm(total=num_epochs, disable=not show_progress, desc='Training', unit='epoch') as pbar:
            for epoch in range(num_epochs):
                epoch_losses = [0.0] * num_losses
                for batch_x, batch_c, _ in dataloader:
                    batch_x = batch_x.to(self.get_device())
                    for loss_id in range(num_losses):
                        if cs is None:
                            new_loss = losses[loss_id].step(batch_x)
                        else:
                            batch_c = batch_c.to(self.get_device())
                            new_loss = losses[loss_id].step(batch_x, batch_c)
                        epoch_losses[loss_id] += new_loss

                avg_epoch_losses_ = map(lambda v: v / len(dataloader), epoch_losses)
                avg_epoch_losses = map(lambda v: "{:.4f}".format(v), avg_epoch_losses_)

                # store the loss
                str_loss = " ".join(map(str, avg_epoch_losses))

                # Update progress bar
                pbar.set_postfix({'loss': str_loss})
                pbar.update(1)

    @classmethod
    def save_model(cls, model, file_path, compression=False):
        """Save the model to the specified file path."""
        file_path = os.path.abspath(file_path)

        model.eval()
        if compression:
            with gzip.open(file_path, 'wb') as pickle_file:
                pickle.dump(model, pickle_file)
        else:
            with open(file_path, 'wb') as pickle_file:
                pickle.dump(model, pickle_file)

        print(f'Model saved to {file_path}')

    @classmethod
    def load_model(cls, file_path):
        """Load the model from the specified file path and return an instance."""
        print(f'Model loaded from {file_path}')

        file_path = os.path.abspath(file_path)
        if file_path.endswith('gz'):
            with gzip.open(file_path, 'rb') as pickle_file:
                model = pickle.load(pickle_file)
        else:
            with open(file_path, 'rb') as pickle_file:
                model = pickle.load(pickle_file)
        
        return model

        
EXAMPLE_RUN = (
    "example run: SURE --help"
)

def parse_args():
    parser = argparse.ArgumentParser(
        description="SURE\n{}".format(EXAMPLE_RUN))

    parser.add_argument(
        "--cuda", action="store_true", help="use GPU(s) to speed up training"
    )
    parser.add_argument(
        "--jit", action="store_true", help="use PyTorch jit to speed up training"
    )
    parser.add_argument(
        "-n", "--num-epochs", default=200, type=int, help="number of epochs to run"
    )
    parser.add_argument(
        "-enum",
        "--enum-discrete",
        default="parallel",
        help="parallel, sequential or none. uses parallel enumeration by default",
    )
    parser.add_argument(
        "-data",
        "--data-file",
        default=None,
        type=str,
        help="the data file",
    )
    parser.add_argument(
        "-cf",
        "--covariate-file",
        default=None,
        type=str,
        help="the file for the record of covariates",
    )
    parser.add_argument(
        "-bs",
        "--batch-size",
        default=1000,
        type=int,
        help="number of cells to be considered in a batch",
    )
    parser.add_argument(
        "-lr",
        "--learning-rate",
        default=0.0001,
        type=float,
        help="learning rate for Adam optimizer",
    )
    parser.add_argument(
        "-zd",
        "--z-dim",
        default=10,
        type=int,
        help="size of the tensor representing the latent variable z variable",
    )
    parser.add_argument(
        "-likeli",
        "--likelihood",
        default='negbinomial',
        type=str,
        choices=['negbinomial', 'multinomial', 'poisson', 'bernoulli'],
        help="specify the distribution likelihood function",
    )
    parser.add_argument(
        "-zi",
        "--zeroinflate",
        action="store_true",
        help="use zeroinflation",
    )
    parser.add_argument(
        "-id",
        "--inverse-dispersion",
        default=10.0,
        type=float,
        help="inverse dispersion prior for negative binomial",
    )
    parser.add_argument(
        "-hl",
        "--hidden-layers",
        nargs="+",
        default=[500],
        type=int,
        help="a tuple (or list) of MLP layers to be used in the neural networks "
        "representing the parameters of the distributions in our model",
    )
    parser.add_argument(
        "-hla",
        "--hidden-layer-activation",
        default='relu',
        type=str,
        choices=['relu','softplus','leakyrelu','linear'],
        help="activation function for hidden layers",
    )
    parser.add_argument(
        "-plf",
        "--post-layer-function",
        nargs="+",
        default=['layernorm'],
        type=str,
        help="post functions for hidden layers, could be none, dropout, layernorm, batchnorm, or combination, default is 'dropout layernorm'",
    )
    parser.add_argument(
        "-paf",
        "--post-activation-function",
        nargs="+",
        default=['none'],
        type=str,
        help="post functions for activation layers, could be none or dropout, default is 'none'",
    )
    parser.add_argument(
        "-dr",
        "--decay-rate",
        default=0.9,
        type=float,
        help="decay rate for Adam optimizer",
    )
    parser.add_argument(
        "--layer-dropout-rate",
        default=0.1,
        type=float,
        help="droput rate for neural networks",
    )
    parser.add_argument(
        "-b1",
        "--beta-1",
        default=0.95,
        type=float,
        help="beta-1 parameter for Adam optimizer",
    )
    parser.add_argument(
        "-64",
        "--float64",
        action="store_true",
        help="use double float precision",
    )
    parser.add_argument(
        "--seed",
        default=None,
        type=int,
        help="seed for controlling randomness in this example",
    )
    parser.add_argument(
        "--save-model",
        default=None,
        type=str,
        help="path to save model for prediction",
    )
    args = parser.parse_args()
    return args

def main():
    args = parse_args()
    assert (
        (args.data_file is not None) and (
            os.path.exists(args.data_file))
    ), "data file must be provided"

    if args.seed is not None:
        set_random_seed(args.seed)

    if args.float64:
        dtype = torch.float64
        torch.set_default_dtype(torch.float64)
    else:
        dtype = torch.float32
        torch.set_default_dtype(torch.float32)

    xs = dt.fread(file=args.data_file, header=True).to_numpy()
    us = None 
    if args.covariate_file is not None:
        us = dt.fread(file=args.covariate_file, header=True).to_numpy()

    input_dim = xs.shape[1]
    covariate_size = 0 if us is None else us.shape[1]

    ###########################################
    sure = SUREVanilla(
        input_dim=input_dim,
        covariate_size=covariate_size,
        dispersion=args.dispersion,
        z_dim=args.z_dim,
        hidden_layers=args.hidden_layers,
        hidden_layer_activation=args.hidden_layer_activation,
        use_cuda=args.cuda,
        config_enum=args.enum_discrete,
        use_zeroinflate=args.zeroinflate,
        loss_func=args.likelihood,
        nn_dropout=args.layer_dropout_rate,
        post_layer_fct=args.post_layer_function,
        post_act_fct=args.post_activation_function,
        dtype=dtype,
    )

    sure.fit(xs, us=us, 
             num_epochs=args.num_epochs,
             learning_rate=args.learning_rate,
             batch_size=args.batch_size,
             beta_1=args.beta_1,
             decay_rate=args.decay_rate,
             use_jax=args.jit,
             config_enum=args.enum_discrete,
             )

    if args.save_model is not None:
        if args.save_model.endswith('gz'):
            SUREVanilla.save_model(sure, args.save_model, compression=True)
        else:
            SUREVanilla.save_model(sure, args.save_model)
    


if __name__ == "__main__":

    main()


import re
import numpy as np
import pandas as pd
from numba import njit
from itertools import chain
from joblib import Parallel, delayed
from typing import Literal

class LabelMatrix:
    def __init__(self):
        self.labels_ = None
        self.control_label = None 
        self.sep_pattern = None
    
    def fit_transform(self, labels, control_label=None, sep_pattern=r'[,;_\s]', speedup: Literal['none','vectorize','parallel']='none'):
        if speedup=='none':
            mat, self.labels_ = label_to_matrix(labels=labels, sep_pattern=sep_pattern)
        elif speedup=='vectorize':
            mat, self.labels_ = vectorized_label_to_matrix(labels=labels, sep_pattern=sep_pattern)
        elif speedup=='parallel':
            mat, self.labels_ = parallel_label_to_matrix(labels=labels, sep_pattern=sep_pattern)
            
        self.labels_ = np.array(self.labels_)
        
        if control_label is not None:
            idx = np.where(self.labels_==control_label)[0]
            mat = np.delete(mat, idx, axis=1)
            self.labels_ = np.delete(self.labels_, idx)
            
        self.control_label = control_label
        self.sep_pattern=sep_pattern
        
        return mat
    
    def transform(self, labels, speedup: Literal['none','vectorize','parallel']='none'):
        sep_pattern = self.sep_pattern
        if speedup=='none':
            mat, labels_ = label_to_matrix(labels=labels, sep_pattern=sep_pattern)
        elif speedup=='vectorize':
            mat, labels_ = vectorized_label_to_matrix(labels=labels, sep_pattern=sep_pattern)
        elif speedup=='parallel':
            mat, labels_ = parallel_label_to_matrix(labels=labels, sep_pattern=sep_pattern)
          
        mat_df = pd.DataFrame(mat, columns=labels_)
        
        labels_valid = [x for x in labels_ if x in self.labels_]  
        mat_df = mat_df[labels_valid]
        
        mat_valid = np.zeros([mat.shape[0], len(self.labels_)])
        mat_valid_df = pd.DataFrame(mat_valid, columns=self.labels_)
        mat_valid_df[labels_valid] = mat_df 
        
        return mat_valid_df.values
        
    def inverse_transform(self, matrix):
        return matrix_to_labels(matrix=matrix, unique_labels=self.labels_)
        
def label_to_matrix(labels, sep_pattern=r'[;_\-\s]'):
    """
    将混合分隔符的多标签数据转换为 0-1 矩阵
    
    Args:
        labels: 原始标签列表，如 ["cat", "dog", "cat;dog", "bird_dog"]
        sep_pattern: 多标签分隔符的正则模式（默认匹配 ; _ - 和空格）
    
    Returns:
        one_hot_matrix: 0-1 矩阵
        unique_labels: 唯一标签列表
    """
    # 统一分隔符
    labels_unified = [re.sub(sep_pattern, ';', label) for label in labels]
    
    # 获取所有唯一标签
    all_unique_labels = sorted(set(chain(*[label.split(';') for label in labels_unified])))
    
    # 生成 0-1 矩阵
    matrix = np.zeros((len(labels), len(all_unique_labels)), dtype=int)
    label_to_idx = {label: i for i, label in enumerate(all_unique_labels)}
    
    for i, label in enumerate(labels_unified):
        for sub_label in label.split(';'):
            if sub_label in label_to_idx:
                matrix[i, label_to_idx[sub_label]] = 1
    
    return matrix, all_unique_labels


def vectorized_label_to_matrix(labels, sep_pattern=r'[;_\-\s]'):
    labels_unified = [re.sub(sep_pattern, ';', label) for label in labels]
    unique_labels = sorted(set(chain(*[label.split(';') for label in labels_unified])))
    
    # 向量化操作
    label_matrix = np.array([label.split(';') for label in labels_unified], dtype=object)
    matrix = np.zeros((len(labels), len(unique_labels)), dtype=int)
    
    for i, label in enumerate(unique_labels):
        matrix[:, i] = np.array([label in lst for lst in label_matrix], dtype=int)
    
    return matrix, unique_labels



def parallel_label_to_matrix(labels, sep_pattern=r'[;_\-\s]', n_jobs=4):
    labels_unified = [re.sub(sep_pattern, ';', label) for label in labels]
    unique_labels = sorted(set(chain(*[label.split(';') for label in labels_unified])))
    
    def process_row(row_labels, unique_labels):
        return [1 if label in row_labels else 0 for label in unique_labels]
    
    label_lists = [label.split(';') for label in labels_unified]
    matrix = Parallel(n_jobs=n_jobs)(
        delayed(process_row)(row, unique_labels) for row in label_lists
    )
    
    return np.array(matrix, dtype=int), unique_labels


def matrix_to_labels(matrix, unique_labels):
    return [';'.join([unique_labels[i] for i in np.where(row)[0]]) 
            for row in matrix]






from .SURE import SURE
from .SURE_vanilla import SUREVanilla
from .SURE_vae import SUREVAE
from .SURE_nsf import SURENF
from .SUREMO import SUREMO

from . import utils 
from . import SURE
from . import SUREMO
from . import SURE_vanilla
from . import SURE_vae
from . import SURE_nsf
from . import atac
from . import dist 
from . import graph

__all__ = ['SURE', 'SURE_vanilla', 'SURE_vae', 'SURE_nsf', 'SUREMO', 'atac', 'utils', 'dist', 'graph']


import numpy as np
import networkx as nx
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.cross_decomposition import PLSRegression
from sklearn.preprocessing import StandardScaler
from scipy.spatial.distance import cdist
from scipy import sparse
import matplotlib.pyplot as plt
from collections import defaultdict, Counter

import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots

from sklearn.neighbors import NearestNeighbors, KDTree
from scipy.sparse import csr_matrix, lil_matrix, diags
from scipy.sparse.linalg import eigs
from scipy.sparse.csgraph import dijkstra
import heapq
from typing import Tuple, List, Dict, Optional, Union
import warnings
warnings.filterwarnings('ignore')




class ClusterNetworkBuilder:
    """
    基于KNN的聚类中心网络构建器
    
    算法步骤：
    1. 建立聚类中心c到数据点x的knn网络H
    2. 基于H建立聚类中心c的网络G
       - 如果两个聚类中心c1可以经过某个数据点到达c2，则c1和c2建立边
       - 边的权重由中介数据点的数量决定
       - 中介数据点的比重越大，边的权重越大
    """
    
    def __init__(self, 
                 k_data_to_centers=5,      # 每个数据点连接到多少个聚类中心
                 k_centers_to_data=10,     # 每个聚类中心连接到多少个数据点
                 weight_method='count',     # 权重计算方法: 'count', 'proportion', 'weighted'
                 min_edge_weight=0.01,     # 最小边权重
                 normalize_weights=True,   # 是否归一化权重
                 symmetric=True,           # 是否创建对称边
                 use_jaccard=False):       # 是否使用Jaccard相似度
        """
        参数:
        k_data_to_centers: 每个数据点连接到k个最近的聚类中心
        k_centers_to_data: 每个聚类中心连接到k个最近的数据点
        weight_method: 权重计算方法
            - 'count': 使用中介数据点的数量
            - 'proportion': 使用中介数据点的比例
            - 'weighted': 使用加权计数
        """
        self.k_data_to_centers = k_data_to_centers
        self.k_centers_to_data = k_centers_to_data
        self.weight_method = weight_method
        self.min_edge_weight = min_edge_weight
        self.normalize_weights = normalize_weights
        self.symmetric = symmetric
        self.use_jaccard = use_jaccard
        
    def build_network(self, x, c, metric='euclidean', return_details=False):
        """
        构建聚类中心网络
        
        参数:
        x: 数据点 (n_samples, n_features)
        c: 聚类中心 (n_clusters, n_features)
        metric: 距离度量
        return_details: 是否返回详细中间结果
        
        返回:
        G: 聚类中心网络
        如果return_details=True，还返回(H, edge_weights_dict, data_assignments)
        """
        n_samples = x.shape[0]
        n_clusters = c.shape[0]
        
        # 步骤1: 构建KNN网络H（二部图）
        H, data_assignments = self._build_knn_bipartite(x, c, metric)
        
        # 步骤2: 基于H构建聚类中心网络G
        G, edge_weights_dict = self._build_cluster_network_from_bipartite(
            H, n_clusters, n_samples
        )
        
        # 为节点添加属性
        self._add_node_attributes(G, c, data_assignments, n_samples)
        
        if return_details:
            return G, H, edge_weights_dict, data_assignments
        return G
    
    def _build_knn_bipartite(self, x, c, metric):
        """构建数据点和聚类中心的KNN二部图H"""
        n_samples = x.shape[0]
        n_clusters = c.shape[0]
        
        # 创建二部图
        H = nx.Graph()
        
        # 添加节点
        # 数据点节点: 前缀'd_'，如'd_0', 'd_1', ...
        for i in range(n_samples):
            H.add_node(f'd_{i}', type='data', index=i, features=x[i])
        
        # 聚类中心节点: 前缀'c_'，如'c_0', 'c_1', ...
        for j in range(n_clusters):
            H.add_node(f'c_{j}', type='cluster', index=j, center=c[j])
        
        # 1. 数据点到聚类中心的连接
        if self.k_data_to_centers > 0:
            # 计算每个数据点到所有聚类中心的距离
            data_to_center_dists = cdist(x, c, metric=metric)
            
            # 为每个数据点找到k个最近的聚类中心
            k1 = min(self.k_data_to_centers, n_clusters)
            for i in range(n_samples):
                # 找到最近的k1个聚类中心
                if k1 < n_clusters:
                    # 使用argpartition提高效率
                    nearest = np.argpartition(data_to_center_dists[i], k1)[:k1]
                else:
                    nearest = np.arange(n_clusters)
                
                for j in nearest:
                    distance = data_to_center_dists[i, j]
                    H.add_edge(f'd_{i}', f'c_{j}', 
                              #weight=1.0/(1.0 + distance),
                              weight=np.exp(-distance),
                              distance=distance,
                              type='data_to_center')
        
        # 2. 聚类中心到数据点的连接
        if self.k_centers_to_data > 0:
            # 计算每个聚类中心到所有数据点的距离
            center_to_data_dists = cdist(c, x, metric=metric)
            
            # 为每个聚类中心找到k个最近的数据点
            k2 = min(self.k_centers_to_data, n_samples)
            for j in range(n_clusters):
                if k2 < n_samples:
                    nearest = np.argpartition(center_to_data_dists[j], k2)[:k2]
                else:
                    nearest = np.arange(n_samples)
                
                for i in nearest:
                    distance = center_to_data_dists[j, i]
                    H.add_edge(f'c_{j}', f'd_{i}',
                              #weight=1.0/(1.0 + distance),
                              weight=np.exp(-distance),
                              distance=distance,
                              type='center_to_data')
        
        # 创建数据点分配矩阵
        data_assignments = self._create_data_assignments(H, n_samples, n_clusters)
        
        return H, data_assignments
    
    def _create_data_assignments(self, H, n_samples, n_clusters):
        """从二部图创建数据点分配矩阵"""
        assignments = np.zeros((n_samples, n_clusters))
        
        for i in range(n_samples):
            data_node = f'd_{i}'
            if data_node in H:
                # 获取连接到这个数据点的所有聚类中心
                cluster_neighbors = [n for n in H.neighbors(data_node) 
                                   if n.startswith('c_')]
                
                for cluster_node in cluster_neighbors:
                    j = int(cluster_node.split('_')[1])
                    # 使用边的权重作为分配权重
                    weight = H[data_node][cluster_node].get('weight', 1.0)
                    assignments[i, j] = weight
        
        # 归一化每行的分配（使每行和为1）
        row_sums = assignments.sum(axis=1, keepdims=True)
        row_sums[row_sums == 0] = 1  # 避免除零
        assignments = assignments / row_sums
        
        return assignments
    
    def _build_cluster_network_from_bipartite(self, H, n_clusters, n_samples):
        """从二部图构建聚类中心网络"""
        G = nx.Graph()
        
        # 添加聚类中心节点
        for j in range(n_clusters):
            G.add_node(j, index=j)
        
        # 计算边权重
        edge_weights = defaultdict(float)
        
        # 方法1: 通过数据点连接聚类中心
        for i in range(n_samples):
            data_node = f'd_{i}'
            if data_node in H:
                # 获取连接到这个数据点的所有聚类中心
                cluster_neighbors = [n for n in H.neighbors(data_node) 
                                   if n.startswith('c_')]
                cluster_indices = [int(n.split('_')[1]) for n in cluster_neighbors]
                
                # 为每对聚类中心添加权重
                for idx1, c1 in enumerate(cluster_indices):
                    for c2 in cluster_indices[idx1+1:]:
                        # 这个数据点同时连接到c1和c2
                        edge_key = tuple(sorted((c1, c2)))
                        
                        # 计算这个数据点对这条边的贡献
                        weight_c1 = H[data_node][f'c_{c1}'].get('weight', 1.0)
                        weight_c2 = H[data_node][f'c_{c2}'].get('weight', 1.0)
                        
                        if self.weight_method == 'count':
                            # 计数：每个中介数据点贡献1
                            contribution = 1.0
                        elif self.weight_method == 'weighted':
                            # 加权：使用连接权重的乘积
                            contribution = weight_c1 * weight_c2
                        else:  # 'proportion'
                            # 比例：平均权重
                            contribution = (weight_c1 + weight_c2) / 2.0
                        
                        edge_weights[edge_key] += contribution
        
        # 如果使用Jaccard相似度
        if self.use_jaccard:
            edge_weights = self._compute_jaccard_weights(H, n_clusters, edge_weights)
        
        # 归一化权重
        if self.normalize_weights and edge_weights:
            max_weight = max(edge_weights.values())
            if max_weight > 0:
                for key in edge_weights:
                    edge_weights[key] /= max_weight
        
        # 添加边到网络G
        for (c1, c2), weight in edge_weights.items():
            if weight >= self.min_edge_weight:
                G.add_edge(c1, c2, weight=weight, 
                          n_mediators=int(weight * 10))  # 估计的中介点数
        
        return G, dict(edge_weights)
    
    def _compute_jaccard_weights(self, H, n_clusters, edge_weights):
        """计算Jaccard相似度作为边权重"""
        # 计算每个聚类中心连接的数据点集合
        cluster_data_sets = {}
        for j in range(n_clusters):
            cluster_node = f'c_{j}'
            if cluster_node in H:
                data_neighbors = {n for n in H.neighbors(cluster_node) 
                                if n.startswith('d_')}
                cluster_data_sets[j] = data_neighbors
        
        # 重新计算边权重为Jaccard相似度
        jaccard_weights = {}
        clusters = list(cluster_data_sets.keys())
        
        for i in range(len(clusters)):
            for j in range(i+1, len(clusters)):
                c1, c2 = clusters[i], clusters[j]
                set1 = cluster_data_sets.get(c1, set())
                set2 = cluster_data_sets.get(c2, set())
                
                if set1 and set2:
                    intersection = len(set1.intersection(set2))
                    union = len(set1.union(set2))
                    jaccard = intersection / union if union > 0 else 0
                    
                    if jaccard > 0:
                        jaccard_weights[(c1, c2)] = jaccard
        
        return jaccard_weights
    
    def _add_node_attributes(self, G, c, data_assignments, n_samples):
        """为节点添加属性"""
        n_clusters = c.shape[0]
        
        for j in range(n_clusters):
            if j in G:
                # 计算分配给这个聚类的数据点
                cluster_assignments = data_assignments[:, j]
                
                # 硬分配计数（分配权重大于0.5）
                hard_count = np.sum(cluster_assignments > 0.5)
                
                # 软分配权重和
                soft_weight = np.sum(cluster_assignments)
                
                # 更新节点属性
                G.nodes[j].update({
                    'center': c[j],
                    'hard_assigned_count': hard_count,
                    'soft_assigned_weight': soft_weight,
                    'proportion': soft_weight / n_samples,
                    'avg_assignment': np.mean(cluster_assignments[cluster_assignments > 0])
                })
    



class KnnClusterNetworkBuilder:
    """
    基于KNN的聚类中心网络构建器
    
    算法：
    1. 建立聚类中心c到数据点x的knn网络H
    2. 建立数据x的knn网络M
    3. 基于H建立聚类中心c的网络G，对任何两个中心c1和c2，
       它们之间的距离定义为从其中一个中心出发经过网络M到达另一个中心的最短距离d，
       c1和c2之间的边的权重定义为exp(-d/t)，t为temperature
    """
    
    def __init__(self, 
                 k_data: int = 5,
                 k_centers: int = 15,
                 temperature: float = 1.0,
                 metric: str = 'euclidean',
                 algorithm: str = 'auto',
                 directed: bool = False,
                 weight_strategy: str = 'distance',
                 symmetric: bool = True,
                 use_sparse: bool = True):
        """
        参数:
        k_data: 数据点KNN的k值
        k_centers: 聚类中心到数据点的KNN的k值
        temperature: 距离到权重的转换温度
        metric: 距离度量 ('euclidean', 'cosine', 'manhattan', 等)
        algorithm: KNN算法 ('auto', 'kd_tree', 'ball_tree', 'brute')
        directed: 是否构建有向图
        weight_strategy: 权重策略 ('distance', 'shared', 'combined')
        symmetric: 是否强制对称（无向图）
        use_sparse: 是否使用稀疏矩阵加速
        """
        self.k_data = k_data
        self.k_centers = k_centers
        self.temperature = temperature
        self.metric = metric
        self.algorithm = algorithm
        self.directed = directed
        self.weight_strategy = weight_strategy
        self.symmetric = symmetric
        self.use_sparse = use_sparse
        
        # 缓存
        self._H = None  # 中心到数据的KNN图
        self._M = None  # 数据的KNN图
        self._G = None  # 最终的中心网络
        self._shortest_paths = None  # 最短路径缓存
        
    def build_network(self, 
                     x: np.ndarray, 
                     c: np.ndarray,
                     return_intermediate: bool = False) -> nx.Graph:
        """
        构建聚类中心网络
        
        参数:
        x: 数据点，形状 (n_samples, n_features)
        c: 聚类中心，形状 (n_clusters, n_features)
        return_intermediate: 是否返回中间图
        
        返回:
        G: 聚类中心网络
        如果return_intermediate为True，也返回(H, M)
        """        
        # 1. 构建数据点的KNN图 M
        M = self._build_data_knn_graph(x)
        
        # 2. 构建聚类中心到数据的KNN图 H
        H = self._build_centers_to_data_knn_graph(x, c)
        
        # 3. 构建最终的聚类中心网络 G
        G = self._build_final_network(x, c, H, M)
        
        # 保存中间结果
        self._M = M
        self._H = H
        self._G = G
        
        if return_intermediate:
            return G, H, M
        return G
    
    def _build_data_knn_graph(self, x: np.ndarray) -> nx.Graph:
        """构建数据点的KNN图 M"""
        n_samples = x.shape[0]
        
        # 构建KNN
        k_actual = min(self.k_data, n_samples - 1)
        if k_actual < 1:
            # 如果样本太少，构建完全图
            M = nx.complete_graph(n_samples)
            # 添加距离作为边权重
            distances = cdist(x, x, self.metric)
            for i in range(n_samples):
                for j in range(i+1, n_samples):
                    M[i][j]['weight'] = distances[i, j]
            return M
        
        # 使用sklearn的KNN
        nbrs = NearestNeighbors(
            n_neighbors=k_actual + 1,  # +1 因为包含自己
            algorithm=self.algorithm,
            metric=self.metric
        ).fit(x)
        
        distances, indices = nbrs.kneighbors(x)
        
        # 构建图
        if self.directed:
            M = nx.DiGraph()
        else:
            M = nx.Graph()
        
        # 添加节点
        M.add_nodes_from(range(n_samples))
        
        # 添加边
        for i in range(n_samples):
            # 获取邻居（排除自己）
            neighbors = indices[i, 1:]  # 第0个是自己
            neighbor_distances = distances[i, 1:]
            
            for j, dist in zip(neighbors, neighbor_distances):
                if not self.directed and i > j:
                    # 对于无向图，避免重复添加
                    if M.has_edge(i, j):
                        continue
                
                # 根据权重策略添加边权重
                if self.weight_strategy == 'distance':
                    weight = dist
                elif self.weight_strategy == 'shared':
                    # 共享邻居数量
                    weight = 1.0
                elif self.weight_strategy == 'combined':
                    weight = dist
                else:
                    weight = 1.0
                
                M.add_edge(i, j, weight=weight, distance=dist)
        
        return M
    
    def _build_centers_to_data_knn_graph(self, x: np.ndarray, c: np.ndarray) -> nx.Graph:
        """构建聚类中心到数据的KNN图 H"""
        n_samples = x.shape[0]
        n_clusters = c.shape[0]
        
        # 构建KNN：从中心到数据
        k_actual = min(self.k_centers, n_samples)
        
        nbrs = NearestNeighbors(
            n_neighbors=k_actual,
            algorithm=self.algorithm,
            metric=self.metric
        ).fit(x)
        
        distances, indices = nbrs.kneighbors(c)
        
        # 构建二分图 H
        H = nx.Graph()
        
        # 添加节点
        # 聚类中心节点：0 到 n_clusters-1
        # 数据点节点：n_clusters 到 n_clusters + n_samples - 1
        H.add_nodes_from(range(n_clusters + n_samples))
        
        # 标记节点类型
        for i in range(n_clusters):
            H.nodes[i]['type'] = 'center'
            H.nodes[i]['center_id'] = i
            H.nodes[i]['original_index'] = i
        
        for i in range(n_samples):
            node_id = n_clusters + i
            H.nodes[node_id]['type'] = 'data'
            H.nodes[node_id]['data_id'] = i
            H.nodes[node_id]['original_index'] = i
        
        # 添加边：中心到数据
        for center_idx in range(n_clusters):
            data_indices = indices[center_idx]
            data_distances = distances[center_idx]
            
            for data_idx, dist in zip(data_indices, data_distances):
                data_node_id = n_clusters + data_idx
                H.add_edge(center_idx, data_node_id, 
                          weight=dist, 
                          distance=dist,
                          center_id=center_idx,
                          data_id=data_idx)
        
        return H
    
    def _build_final_network(self, 
                           x: np.ndarray, 
                           c: np.ndarray,
                           H: nx.Graph,
                           M: nx.Graph) -> nx.Graph:
        """构建最终的聚类中心网络 G"""
        n_clusters = c.shape[0]
        n_samples = x.shape[0]
        
        # 创建最终的网络
        G = nx.Graph()
        
        # 添加聚类中心节点
        for i in range(n_clusters):
            G.add_node(i, center=c[i], index=i)
        
        # 预计算数据点之间的最短路径
        data_shortest_paths = self._compute_data_shortest_paths(M, n_samples)
        
        # 计算中心之间的最短路径距离        
        # 获取H中每个中心连接的数据点
        center_to_data = {}
        for center_idx in range(n_clusters):
            connected_data = []
            for neighbor in H.neighbors(center_idx):
                if H.nodes[neighbor]['type'] == 'data':
                    data_id = H.nodes[neighbor]['data_id']
                    weight = H[center_idx][neighbor]['weight']
                    connected_data.append((data_id, weight))
            center_to_data[center_idx] = connected_data
        
        # 计算所有中心对之间的最短路径
        center_distances = np.full((n_clusters, n_clusters), np.inf)
        center_paths = {}
        
        for i in range(n_clusters):
            center_distances[i, i] = 0
            for j in range(i+1, n_clusters):
                # 计算从中心i到中心j的最短路径
                distance, path = self._compute_center_to_center_distance(
                    i, j, center_to_data, data_shortest_paths, H
                )
                
                if distance < np.inf:
                    center_distances[i, j] = distance
                    center_distances[j, i] = distance
                    center_paths[(i, j)] = path
                    center_paths[(j, i)] = list(reversed(path))
        
        # 添加边到网络G
        edges_added = 0
        
        for i in range(n_clusters):
            for j in range(i+1, n_clusters):
                distance = center_distances[i, j]
                
                if distance < np.inf:
                    # 计算边权重: exp(-d/t)
                    weight = np.exp(-distance / self.temperature)
                    
                    # 获取路径信息
                    path_info = center_paths.get((i, j), [])
                    
                    # 添加边
                    G.add_edge(i, j, 
                              weight=weight,
                              distance=distance,
                              raw_distance=distance,
                              temperature=self.temperature,
                              path=path_info)
                    
                    edges_added += 1
                
        # 如果是对称的，确保权重对称
        if self.symmetric and not self.directed:
            for i, j in G.edges():
                if G[i][j]['weight'] != G[j][i]['weight']:
                    # 取平均值
                    avg_weight = (G[i][j]['weight'] + G[j][i]['weight']) / 2
                    G[i][j]['weight'] = avg_weight
                    G[j][i]['weight'] = avg_weight
        
        return G
    
    def _compute_data_shortest_paths(self, M: nx.Graph, n_samples: int) -> np.ndarray:
        """预计算数据点之间的最短路径距离"""
        
        if self.use_sparse and n_samples > 100:
            # 使用稀疏矩阵和Dijkstra算法
            # 创建邻接矩阵
            adj_matrix = np.full((n_samples, n_samples), np.inf)
            np.fill_diagonal(adj_matrix, 0)
            
            for u, v, data in M.edges(data=True):
                weight = data.get('weight', 1.0)
                adj_matrix[u, v] = weight
                if not self.directed:
                    adj_matrix[v, u] = weight
            
            # 使用Dijkstra算法
            from scipy.sparse import csr_matrix
            from scipy.sparse.csgraph import dijkstra
            
            sparse_adj = csr_matrix(adj_matrix)
            distances = dijkstra(sparse_adj, directed=self.directed)
            
            return distances
        else:
            # 使用NetworkX的最短路径
            distances = np.full((n_samples, n_samples), np.inf)
            np.fill_diagonal(distances, 0)
            
            # 对于小图，使用所有对最短路径
            if n_samples < 1000:
                all_pairs = dict(nx.all_pairs_dijkstra_path_length(M, weight='weight'))
                for u in all_pairs:
                    for v, dist in all_pairs[u].items():
                        distances[u, v] = dist
            else:
                # 对于大图，只计算必要的
                warnings.warn("数据点太多，最短路径计算可能较慢")
                return None
            
            return distances
    
    def _compute_center_to_center_distance(self,
                                         center_i: int,
                                         center_j: int,
                                         center_to_data: Dict[int, List[Tuple[int, float]]],
                                         data_shortest_paths: Optional[np.ndarray],
                                         H: nx.Graph) -> Tuple[float, List]:
        """
        计算从中心i到中心j的最短路径距离
        
        路径格式: 中心i -> 数据a -> ... -> 数据b -> 中心j
        """
        data_i_list = center_to_data[center_i]  # [(data_id, weight), ...]
        data_j_list = center_to_data[center_j]  # [(data_id, weight), ...]
        
        if not data_i_list or not data_j_list:
            return np.inf, []
        
        min_distance = np.inf
        min_path = []
        
        # 遍历所有可能的起点和终点数据点
        for data_i, weight_i in data_i_list:
            for data_j, weight_j in data_j_list:
                if data_i == data_j:
                    # 同一个数据点
                    total_distance = weight_i + weight_j
                    path = [center_i, data_i, center_j]
                else:
                    # 需要计算数据点之间的最短路径
                    if data_shortest_paths is not None:
                        data_distance = data_shortest_paths[data_i, data_j]
                    else:
                        # 实时计算最短路径
                        try:
                            data_distance = nx.shortest_path_length(
                                self._M, data_i, data_j, weight='weight'
                            )
                        except nx.NetworkXNoPath:
                            continue
                    
                    if data_distance < np.inf:
                        total_distance = weight_i + data_distance + weight_j
                        path = [center_i, data_i, data_j, center_j]
                    else:
                        continue
                
                if total_distance < min_distance:
                    min_distance = total_distance
                    min_path = path
        
        return min_distance, min_path
    
    


class EdgeAllocation:
    """
    使用偏最小二乘回归(PLSR)将数据点分配到图边上的类
    """
    
    def __init__(self, n_components: int = 2, alpha: float = 0.1):
        """
        初始化
        
        参数:
        n_components: PLS回归的组件数
        alpha: 分配的平滑参数，防止过度集中在最近的点
        """
        self.n_components = n_components
        self.alpha = alpha
        self.pls = None
        self.scaler_x = StandardScaler()
        self.scaler_y = StandardScaler()
        
    def _point_to_line_distance(self, point: np.ndarray, 
                               start: np.ndarray, 
                               end: np.ndarray) -> float:
        """
        计算点到线段的距离
        
        参数:
        point: 数据点坐标
        start: 线段起点
        end: 线段终点
        
        返回:
        点到线段的距离
        """
        # 线段向量
        line_vec = end - start
        point_vec = point - start
        
        # 线段长度
        line_len = np.linalg.norm(line_vec)
        
        if line_len == 0:
            return np.linalg.norm(point - start)
        
        # 归一化的线段向量
        line_unit = line_vec / line_len
        
        # 点在线段上的投影长度
        proj_length = np.dot(point_vec, line_unit)
        
        if proj_length < 0:
            # 投影在线段起点之前
            return np.linalg.norm(point - start)
        elif proj_length > line_len:
            # 投影在线段终点之后
            return np.linalg.norm(point - end)
        else:
            # 投影在线段上
            projection = start + proj_length * line_unit
            return np.linalg.norm(point - projection)
    
    def fit_transform(self, x: np.ndarray, c: np.ndarray, G: nx.Graph, metric: str='cosine') -> Dict[Tuple, float]:
        """
        将数据点分配到图的边上，返回边的权重（分配到的数据比例）
        
        参数:
        x: 数据点，形状 (n_samples, n_features)
        c: 聚类中心，形状 (n_clusters, n_features)
        G: 图，节点为聚类中心索引，边表示连接
        
        返回:
        边的权重字典，键为边(节点i, 节点j)，值为分配的权重
        """
        n_samples, n_features = x.shape
        n_clusters = c.shape[0]
        
        # 1. 准备PLS回归数据
        # 为每个数据点创建聚类中心的软分配标签
        # 使用距离的倒数作为软标签
        distances = cdist(x, c, metric=metric)
        #soft_labels = 1.0 / (distances + 1e-10)  # 防止除零
        soft_labels = np.exp(-distances)
        soft_labels = soft_labels / soft_labels.sum(axis=1, keepdims=True)  # 归一化
        
        # 2. 拟合PLS回归模型
        x_scaled = self.scaler_x.fit_transform(x)
        y_scaled = self.scaler_y.fit_transform(soft_labels)
        
        self.pls = PLSRegression(n_components=min(self.n_components, n_features))
        self.pls.fit(x_scaled, y_scaled)
        
        # 3. 预测每个点对聚类中心的归属度
        x_scaled = self.scaler_x.transform(x)
        y_pred = self.pls.predict(x_scaled)
        y_pred = np.maximum(y_pred, 0)  # 确保非负
        y_pred = y_pred / y_pred.sum(axis=1, keepdims=True)  # 归一化
        
        # 4. 将点分配到边上
        edges = list(G.edges())
        edge_weights = {edge: 0.0 for edge in edges}
        
        for i in range(n_samples):
            point = x[i]
            cluster_probs = y_pred[i]  # 点属于每个聚类中心的概率
            
            # 对每个边计算分配权重
            edge_distances = []
            edge_probs = []
            
            for edge in edges:
                node_i, node_j = edge
                # 获取聚类中心坐标
                center_i = c[node_i]
                center_j = c[node_j]
                
                # 计算点到边的距离
                dist = self._point_to_line_distance(point, center_i, center_j)
                
                # 计算边相关的聚类中心的平均概率
                edge_prob = (cluster_probs[node_i] + cluster_probs[node_j]) / 2
                
                edge_distances.append(dist)
                edge_probs.append(edge_prob)
            
            # 将距离转换为相似度（使用高斯核）
            edge_distances = np.array(edge_distances)
            edge_probs = np.array(edge_probs)
            
            # 计算边对点的吸引力
            # 结合距离相似度和聚类概率
            similarities = np.exp(-self.alpha * edge_distances) * edge_probs
            
            # 归一化为分配权重
            if similarities.sum() > 0:
                edge_weights_for_point = similarities / similarities.sum()
            else:
                edge_weights_for_point = np.ones(len(edges)) / len(edges)
            
            # 累加权重
            for (edge, weight) in zip(edges, edge_weights_for_point):
                edge_weights[edge] += weight / n_samples  # 平均到每个点
        
        return edge_weights
    

def visualize_network(G, c=None, node_size_attr='weight', node_size_scale=5000, edge_size_scale=30, figsize=(12, 10)):
    """可视化聚类中心网络"""
    fig, ax1 = plt.subplots(1, 1, figsize=figsize)
    
    # 位置布局（如果提供了聚类中心坐标，使用它们；否则使用力导向布局）
    if c is not None and c.shape[1] >= 2:
        # 使用前两个维度作为位置
        pos = {i: (c[i, 0], c[i, 1]) for i in G.nodes()}
        
        # 如果只有1D，添加一个虚拟维度
        if c.shape[1] == 1:
            pos = {i: (c[i, 0], i) for i in G.nodes()}
    else:
        pos = nx.spring_layout(G, seed=42, weight='weight')
    
    # 节点大小基于权重
    node_weights = [G.nodes[node][node_size_attr] for node in G.nodes()]
    node_sizes = [node_size_scale * w for w in node_weights]
    
    # 节点颜色基于权重
    node_colors = node_weights
    
    # 边宽度基于权重
    edge_weights = [G[u][v]['weight'] for u, v in G.edges()]
    edge_widths = [edge_size_scale * w for w in edge_weights]
    
    # 绘制网络
    nodes = nx.draw_networkx_nodes(G, pos, 
                                  node_size=node_sizes,
                                  node_color=node_colors,
                                  cmap='viridis',
                                  alpha=0.8,
                                  edgecolors='black',
                                  linewidths=1,
                                  ax=ax1)
    
    nx.draw_networkx_edges(G, pos,
                          width=edge_widths,
                          alpha=0.6,
                          edge_color='gray',
                          ax=ax1)
    
    nx.draw_networkx_labels(G, pos,
                           font_size=10,
                           font_weight='bold',
                           ax=ax1)
    
    # 添加颜色条
    sm = plt.cm.ScalarMappable(cmap='viridis', 
                               norm=plt.Normalize(vmin=min(node_colors), 
                                                  vmax=max(node_colors)))
    sm.set_array([])
    cbar = plt.colorbar(sm, ax=ax1, shrink=0.8)
    cbar.set_label('Node Weight')
    
    ax1.axis('off')
    
    plt.tight_layout()
    plt.show()




def visualize_network_interactive(G, c=None, node_size_scale=5000, edge_size_scale=30, 
                                 figsize=(1000, 800), show_labels=True, title=""):
    """
    使用Plotly的交互式可视化聚类中心网络
    
    参数:
    G: NetworkX图
    c: 聚类中心坐标 (n_clusters, n_features)
    node_size_scale: 节点大小缩放因子
    edge_size_scale: 边宽度缩放因子
    figsize: 图形大小 (宽度, 高度)
    show_labels: 是否显示节点标签
    title: 图形标题
    
    返回:
    Plotly图形对象
    """
    # 1. 计算节点位置
    if c is not None and c.shape[1] >= 2:
        # 使用前两个维度作为位置
        pos = {i: (float(c[i, 0]), float(c[i, 1])) for i in G.nodes()}
        
        # 如果只有1D，添加一个虚拟维度
        if c.shape[1] == 1:
            pos = {i: (float(c[i, 0]), float(i)) for i in G.nodes()}
    else:
        pos = nx.spring_layout(G, seed=42, weight='weight')
    
    # 2. 准备节点数据
    node_x = []
    node_y = []
    node_sizes = []
    node_colors = []
    node_labels = []
    node_hover_texts = []
    node_ids = []
    
    for node in G.nodes():
        x, y = pos[node]
        node_x.append(x)
        node_y.append(y)
        node_ids.append(node)
        
        # 获取节点属性
        node_weight = G.nodes[node].get('weight', 1.0)
        node_sizes.append(node_size_scale * node_weight)
        node_colors.append(node_weight)
        
        # 构建节点标签
        label_parts = [f"Node {node}"]
        label_parts.append(f"Weight: {node_weight:.4f}")
        
        # 添加其他重要属性
        for attr_name, attr_value in G.nodes[node].items():
            if attr_name not in ['weight', 'center', 'index']:
                if isinstance(attr_value, (int, float)):
                    label_parts.append(f"{attr_name}: {attr_value:.4f}")
                else:
                    label_parts.append(f"{attr_name}: {attr_value}")
        
        node_labels.append(f"Node {node}")
        node_hover_texts.append("<br>".join(label_parts))
    
    # 3. 准备边数据
    edge_x = []
    edge_y = []
    edge_widths = []
    edge_colors = []
    edge_hover_texts = []
    
    for u, v in G.edges():
        x0, y0 = pos[u]
        x1, y1 = pos[v]
        
        # 添加边的坐标（包括None用于分隔）
        edge_x.extend([x0, x1, None])
        edge_y.extend([y0, y1, None])
        
        # 获取边属性
        edge_data = G[u][v]
        edge_weight = edge_data.get('weight', 1.0)
        edge_distance = edge_data.get('distance', 0.0)
        
        # 计算边宽度
        width = edge_size_scale * edge_weight
        edge_widths.extend([width, width, width])  # 为每个点重复
        
        # 边颜色
        edge_colors.extend(['rgba(128, 128, 128, 0.7)'] * 3)
        
        # 构建悬停文本
        hover_parts = [f"Edge ({u} ↔ {v})"]
        hover_parts.append(f"Weight: {edge_weight:.4f}")
        if 'distance' in edge_data:
            hover_parts.append(f"Distance: {edge_distance:.4f}")
        
        # 添加其他边属性
        for attr_name, attr_value in edge_data.items():
            if attr_name not in ['weight', 'distance']:
                if isinstance(attr_value, (int, float)):
                    hover_parts.append(f"{attr_name}: {attr_value:.4f}")
                else:
                    hover_parts.append(f"{attr_name}: {attr_value}")
        
        hover_text = "<br>".join(hover_parts)
        edge_hover_texts.extend([hover_text, hover_text, hover_text])
    
    # 4. 创建图形
    fig = go.Figure()
    
    # 添加边轨迹
    fig.add_trace(go.Scatter(
        x=edge_x,
        y=edge_y,
        mode='lines',
        line=dict(
            width=edge_widths[0] if edge_widths else 1,
            color='rgba(128, 128, 128, 0.7)'
        ),
        hoverinfo='text',
        text=edge_hover_texts,
        name='Edges',
        showlegend=False
    ))
    
    # 添加节点轨迹
    fig.add_trace(go.Scatter(
        x=node_x,
        y=node_y,
        mode='markers' + ('+text' if show_labels else ''),
        marker=dict(
            size=node_sizes,
            color=node_colors,
            colorscale='Viridis',
            showscale=True,
            colorbar=dict(
                title="Node Weight",
                thickness=20,
                x=1.02
            ),
            line=dict(color='black', width=2)
        ),
        text=node_labels if show_labels else None,
        hovertext=node_hover_texts,
        hoverinfo='text',
        name='Nodes',
        textposition="top center",
        textfont=dict(size=12, color='black', family="Arial Black")
    ))
    
    # 5. 更新布局
    fig.update_layout(
        title=dict(
            text=title,
            font=dict(size=20, family="Arial Black"),
            x=0.5
        ),
        showlegend=False,
        hovermode='closest',
        width=figsize[0],
        height=figsize[1],
        xaxis=dict(
            showgrid=True,
            zeroline=False,
            showticklabels=False,
            title='X Coordinate' if c is not None else None
        ),
        yaxis=dict(
            showgrid=True,
            zeroline=False,
            showticklabels=False,
            title='Y Coordinate' if c is not None else None
        ),
        margin=dict(l=50, r=50, t=80, b=50),
        plot_bgcolor='white',
        paper_bgcolor='white'
    )
    
    # 6. 添加网络统计信息
    stats_text = f"""
    <b>Network Statistics:</b><br>
    • Nodes: {G.number_of_nodes()}<br>
    • Edges: {G.number_of_edges()}<br>
    • Average Node Weight: {np.mean(node_colors):.4f}<br>
    • Average Edge Weight: {np.mean([G[u][v].get('weight', 1.0) for u, v in G.edges()]) if G.number_of_edges() > 0 else 0:.4f}
    """
    
    fig.add_annotation(
        x=0.02,
        y=0.98,
        xref="paper",
        yref="paper",
        text=stats_text,
        showarrow=False,
        align="left",
        bordercolor="black",
        borderwidth=1,
        borderpad=4,
        bgcolor="white",
        opacity=0.9,
        font=dict(size=12)
    )
    
    return fig






    
from collections import defaultdict

def keep_top_k_edges_per_node(G, k, weight_attr='weight', inplace=False):
    """
    为每个节点保留权重最大的top k条边
    
    参数:
    G: NetworkX图
    k: 每个节点保留的最大边数
    weight_attr: 权重属性名
    inplace: 是否修改原图
    
    返回:
    处理后的图
    """
    if not inplace:
        H = G.copy()
    else:
        H = G
    
    # 如果k大于等于最大度，直接返回
    max_degree = max(dict(H.degree()).values(), default=0)
    if k >= max_degree:
        return H
    
    # 记录要删除的边
    edges_to_remove = set()
    
    for node in H.nodes():
        # 获取该节点的所有边及其权重
        edges = []
        for neighbor in H.neighbors(node):
            weight = H[node][neighbor].get(weight_attr, 1.0)
            edges.append((weight, node, neighbor))
        
        if len(edges) <= k:
            # 如果边数 <= k，保留所有边
            continue
        
        # 按权重排序，保留最大的k条边
        edges.sort(reverse=True, key=lambda x: x[0])
        edges_to_keep = set()
        
        for i in range(min(k, len(edges))):
            _, u, v = edges[i]
            # 确保每条边只被记录一次
            edge_key = (min(u, v), max(u, v)) if not H.is_directed() else (u, v)
            edges_to_keep.add(edge_key)
        
        # 标记要删除的边
        for _, u, v in edges:
            edge_key = (min(u, v), max(u, v)) if not H.is_directed() else (u, v)
            if edge_key not in edges_to_keep:
                edges_to_remove.add(edge_key)
    
    # 删除边
    if H.is_directed():
        edges_to_remove_list = [(u, v) for (u, v) in edges_to_remove]
    else:
        # 无向图需要确保双向都被删除
        edges_to_remove_list = list(edges_to_remove)
    
    H.remove_edges_from(edges_to_remove_list)
        
    return H